#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <malloc.h>
/*Program trazi minimum (ili maksimum) funkcije dobrote sa inicijalnim pocetnim intervalom (dg[i], gg[i])
(kojeg kasnije moze prosiriti metoda Mutacija ili OPTIMIZIRAN_KOD_Mutacija_i_Rekombinacija).
Svaki vektor ima D parametara. Np je velicina populacije. F je mutacijski faktor. CR je vjerojatnost rekombinacije.

Parametri prvog vektora se nalaze od indeksa 0 do indeksa D-1.
Parametri drugog vektora se nalaze od indeksa D do indeksa 2*D-1.
...
Parametri Np-og vektora se nalaze od indeksa (Np-1)*D do indeksa Np*D-1.
*/

//funkcija se napise tu, redak nize
#define funkcija x[0]*x[0]
/*primjeri funkcija:
x[0]*x[0] // y = f(x) = x^2
(pow(sin(x[0]),2)+3*cos(x[0])-4)*sin(3*x[0]) // y = f(x) = (sin(x)*sin(x)+3*cos(x)-4)*sin(3*x)
sin(x[0])/x[0] // y = f(x) = sinc(x)
*/

const char ulazna[] = "ulazna.txt";
const char izlazna[] = "ispis.txt";
const char format[] = " %13.10lf";
const char format_za_dobrotu[] = "\t%12.9lf"; //" %13.10lf";

/* u komentarima pise sto pojedini macro radi u programu ako je omogucen
i sto radi ako nije omogucen (tj. ako je zakomentiran) poslije rijeci "inace"
*/
#define ISPIS_U_DAT		//sve zapisuje u izlaznu datoteku (datoteku definiranu pod const char izlazna[]), inace ispisuje u stdout
#define ISPIS_ZA_GNUPLOT	//ispisuje u formatu (broj_generacije najbolja_jedinka) koji gnuplot moze koristiti, inace ispise sve relevantne podatke
#define PRIKAZ_DOBROTE		//uz parametre vektora ispise vrijednost funkcije dobrote, inace ju ne ispise
#define OPTIMIZIRAN_KOD		//ne koristi vektor v i smanjuje broj evaluacija do zavrsetka programa, inace radi kao u teoriji (malo neoptimizirano)
//#define CMD			//prima argumente preko komandne linije, inace ih cita iz ulazne datoteke (datoteke definirane pod const char ulazna[])

typedef double tip;
void ispis_r(int *r, int i, FILE *di);

void fatal (char *poruka) {
	fputs (poruka, stderr);
	fputs ("\n", stderr);
#ifdef _DEBUG
	system("pause");
#endif
	exit (1);
}

tip fja_dobrote(tip *x, unsigned D) {
	return funkcija;
}

int compare( tip *arg1, tip *arg2, unsigned D ) { // vraca 1 ako je arg1 > arg2, -1 ako je arg1 < arg2, i 0 ako su isti
	tip rez;
	rez = fja_dobrote(arg1, D) - fja_dobrote(arg2, D);
	if (rez > 0) return 1;
	else if (rez < 0) return -1;
	else return 0;
}

void Inicijalizacija(tip *x, tip *dg, tip *gg, unsigned int D, unsigned int Np){
	unsigned int i, j;
	for (i = 0; i < Np; ++i){
		for (j = 0; j < D; ++j){
			x[i*D+ j] = (tip) (((double)rand() / RAND_MAX) * (gg[j] - dg[j]) + dg[j]);
		}
	}
}

void OPTIMIZIRAN_KOD_Mutacija_i_Rekombinacija(tip *x, tip *u, unsigned int D, unsigned int Np, double F, double CR, FILE *di){
	unsigned int i, j, k, r[3], Irand;
	Irand = (int) (D* ((double)rand() / RAND_MAX) );
	for (i = 0; i < Np; ++i){
		for (k = 0; k < 3; ++k){
			//r[k] = rand() % Np; // ako Np nije veci od RAND_MAX
			r[k] = (int) (Np * (rand() / ((double)RAND_MAX + 1)) );
			for (j = 0; j < k; ++j){
				if (r[k] == r[j]) k--; //r0, r1 i r2 moraju biti medusobno razliciti, pa se ponavlja slucajan odabir
			}
		}
#ifndef ISPIS_ZA_GNUPLOT
		ispis_r(r, i, di);
#endif
		for (j = 0; j < D; ++j){
			if ( ( ((double)rand() / RAND_MAX) <= CR) || (Irand == j) ) //parametar s indeksom Irand se uvijek uzima od v-a
			u[i*D+j] = x[r[0]*D+j] + F * (x[r[1]*D+j] - x[r[2]*D+j]);
			else u[i*D+j] = x[i*D+j];
		}
	}

#ifndef ISPIS_ZA_GNUPLOT
	#ifdef ISPIS_U_DAT
		fprintf(di, "\n");
	#else
		printf("\n");
	#endif
#endif
}

void Mutacija(tip *x, tip *v, unsigned int D, unsigned int Np, double F, FILE *di){
	unsigned int i, j, k, r[3];
	for (i = 0; i < Np; ++i){
		for (k = 0; k < 3; ++k){
			//r[k] = rand() % Np; // ako Np nije veci od RAND_MAX; uobicajeno nije
			r[k] = (int) (Np * (rand() / ((double)RAND_MAX + 1)) );
			for (j = 0; j < k; ++j){
				if (r[k] == r[j]) k--; //r0, r1 i r2 moraju biti medusobno razliciti, pa ponavljamo slucajan odabir
			}
		}
#ifndef ISPIS_ZA_GNUPLOT
		ispis_r(r, i, di);
#endif
		for (j = 0; j < D; ++j){
			v[i*D+j] = x[r[0]*D+j] + F * (x[r[1]*D+j] - x[r[2]*D+j]);
		}
	}

#ifndef ISPIS_ZA_GNUPLOT
	#ifdef ISPIS_U_DAT
		fprintf(di, "\n");
	#else
		printf("\n");
	#endif
#endif
}

void Rekombinacija(tip *x, tip *v, tip *u, unsigned int D, unsigned int Np, double CR){
	unsigned int i, Irand;
	Irand = (int) (D* ((double)rand() / RAND_MAX) );
	for (i = 0; i < D*Np; ++i){
		if ( ((double)rand() / RAND_MAX <= CR) || (Irand == (i%D)) )
			u[i] = v[i]; //parametar s indeksom Irand se uvijek uzima od v-a
		else u[i] = x[i];
	}
}

void Selekcija(tip *x, tip *u, unsigned int D, unsigned int Np, int ekstrem){
	unsigned int i, j;
	for (i = 0; i < Np; ++i) {
	if (compare(&x[i*D], &u[i*D], D) == ekstrem )	// ekstrem = 1, izvrsit ce se ako je x > u, trazimo minimum
													// ekstrem = -1, izvrsit ce se ako je x < u, trazimo maksimum
		for (j = 0; j < D; ++j ){
			x[i*D+j] = u[i*D+j];
		}
	}
}

void ispis_vektora(tip *x, unsigned int D, unsigned int Np, unsigned G, char vektor, FILE *di){
	unsigned int i, j;
#ifdef ISPIS_U_DAT
	fprintf(di, "%u. generacija vektora %c:\n", G, vektor);
#else
	printf("%u. generacija vektora %c:\n", G, vektor);
#endif
	for (i = 0; i < Np; ++i){
#ifdef ISPIS_U_DAT
		fprintf(di, "%d. vektor:", i+1);
#else
		printf("%d. vektor:", i+1);
#endif
		for (j = 0; j < D; ++j){
#ifdef ISPIS_U_DAT
			fprintf(di, format, x[i*D+j]);
#else
			printf(format, x[i*D+j]);
#endif
		}
#ifdef PRIKAZ_DOBROTE
	#ifdef ISPIS_U_DAT
			fprintf(di, format_za_dobrotu, fja_dobrote(&x[i*D], D));
	#else
			printf(format_za_dobrotu, fja_dobrote(&x[i*D], D));
	#endif
#endif
#ifdef ISPIS_U_DAT
		fprintf(di, "\n");
#else
		printf("\n");
#endif
	}

#ifdef ISPIS_U_DAT
	fprintf(di, "\n");
#else
	printf("\n");
#endif
}

#ifdef ISPIS_ZA_GNUPLOT

void ispis_najbolje_jedinke(tip *x, unsigned int D, unsigned int Np, unsigned G, int ekstrem, FILE *di){
	unsigned int i, j;
	tip *best;
	if((best = (tip *)malloc(D * sizeof(tip))) == NULL ) fatal ("Nedovoljno memorije");
	for (i = 0; i < D; ++i) {
		best[i] = x[i];
	}
#ifdef ISPIS_U_DAT
	fprintf(di, "%-3u", G);
#else
	printf("%-3u", G);
#endif
	for (i = 1; i < Np; ++i) {
		if (compare(best, &x[i*D], D) == ekstrem )	// ekstrem = 1, izvrsit ce se ako je best > x[i*D], trazimo minimum
													// ekstrem = -1, izvrsit ce se ako je best < x[i*D], trazimo maksimum
			for (j = 0; j < D; ++j ){
				best[j] = x[i*D+j];
			}
	}
	for (i = 0; i < D; ++i) {
#ifdef ISPIS_U_DAT
		fprintf(di, format, best[i]);
#else
		printf(format, best[i]);
#endif
	}
#ifdef PRIKAZ_DOBROTE
	#ifdef ISPIS_U_DAT
		fprintf(di, format_za_dobrotu, fja_dobrote(&best[0], D));
	#else
		printf(format_za_dobrotu, fja_dobrote(&best[0], D));
	#endif
#endif

#ifdef ISPIS_U_DAT
	fprintf(di, "\n");
#else
	printf("\n");
#endif
	free(best);
}

#else

void ispis_r(int *r, int i, FILE *di){
	int j;
#ifdef ISPIS_U_DAT
	fprintf (di, "%d. vektor v je nacinjen od x-eva:\n", i+1);
#else
	printf ("%d. vektor v je nacinjen od x-eva:\n", i+1);
#endif
	for (j = 0; j < 3; ++j){
#ifdef ISPIS_U_DAT
		fprintf(di, "r%d = %d\n", j+1, r[j]+1);
#else
		printf("r%d = %d\n", j+1, r[j]+1);
#endif
	}
}
#endif

#ifdef CMD
int main (int argn, char *argv[]) {
#else
int main () {
#endif
#ifdef ISPIS_U_DAT
	FILE *di;
#endif
#ifndef CMD
	FILE *du;
#endif
	double *gg, *dg; //gg - gornje granice, dg - donje granice, stop - broj generacija
	unsigned int i, j, D, Np, G = 1, stop; //G je broj trenutne generacije
	double F, CR; //F - mutacijski faktor, CR - konstanta krizanja
	char kod[100];
	tip *x, *u;
#ifndef OPTIMIZIRAN_KOD
	tip *v;
#endif
	int ekstrem; // 1 za minimum, -1 za maksimum
	/*x je ciljni vektor (eng. target vector)
	v je donorski vektor (eng. donor vector)
	u je probni vektor (eng. trial vector)
	*/

	srand( (unsigned)time( NULL ) );

#ifdef CMD
	if (argn < 8) { // 6 + 2*D + 1;za ime datoteke, D >= 1
		fatal("Premalo argumenata\n");
	}
	F = atof(argv[1]);
	//if (_atodbl(&F, argv[1])) fatal("Overflow.\n");
	D = atoi(argv[2]);
	if (D < 1) fatal("Premalo parametara. Mora biti barem jedan parametar.\n");
	if (argn < (6 + 2*D)) fatal("Premalo argumenata\n");
	Np = atoi(argv[3]);
	CR = atof(argv[4]);
	//if (_atodbl(&CR, argv[4])) fatal("Overflow.\n");
	stop = atoi(argv[5]);
	ekstrem = atoi(argv[6]);
	j = 7; // slijedeci ulazni parametar
	if((dg = (double *)malloc(D * sizeof(double))) == NULL ) fatal ("Nedovoljno memorije_dg");
	if((gg = (double *)malloc(D * sizeof(double))) == NULL ) fatal ("Nedovoljno memorije_gg");

	for (i = 0; i < D; ++i) {
		dg[i] = atof(argv[j++]);
		gg[i] = atof(argv[j++]);
		if (dg[i] == gg[i]) fatal("Donja granica i gornja granica su jednake.\n");
	}
#else
	if ((du = fopen (ulazna, "rt")) == NULL) fatal ("Ne mogu otvoriti ulaznu datoteku.");

	while ( fgetc(du) == '#') fgets(kod, 100, du);
	if (fscanf (du, "%s\n", &kod) != 1) fatal ("Krivi zapis.\n");
	while (!(kod[0] == 'E' && kod[1] == 0)) { // D je vec procitan
		while ( fgetc(du) != '\n');
		if ( fgetc(du) == '\n') {
			while ( fgetc(du) == '\n');
			if (fscanf (du, "%s\n", &kod) != 1) fatal ("Krivi zapis.\n");
		}
	}
	if (fscanf (du, "%lf %u %u %lf %u %d\n", &F, &D, &Np, &CR, &stop, &ekstrem) != 6) fatal ("Krivih 6 zapisa.\n");
	if (D < 1) fatal("Premalo parametara. Mora biti barem jedan parametar.\n");
	if((dg = (double *)malloc(D * sizeof(double))) == NULL ) fatal ("Nedovoljno memorije_dg");
	if((gg = (double *)malloc(D * sizeof(double))) == NULL ) fatal ("Nedovoljno memorije_gg");
	for (i = 0; i < D; ++i) {
		if (fscanf (du, "%lf %lf\n", &dg[i], &gg[i]) != 2) fatal ("Krivih 2 zapisa.\n");
		if (dg[i] == gg[i]) fatal("Donja granica i gornja granica su jednake.\n");
	}
	fclose (du);
#endif
	
	if (Np < 4) fatal("Premala populacija Np. Mora biti veca od 3.\n");
	if (F < 0 || F > 2) fatal("Mutacijski faktor F treba biti izmedu 0 i 2.\n");
	if (CR < 0 || CR > 1) fatal("Vjerojatnost krizanja CR treba biti izmedu 0 i 1.\n");
	
	if((x = (tip *)malloc(Np * D * sizeof(tip))) == NULL ) fatal ("Nedovoljno memorije_x");
#ifndef OPTIMIZIRAN_KOD
	if((v = (tip *)malloc(Np * D * sizeof(tip))) == NULL ) fatal ("Nedovoljno memorije_v");
#endif
	if((u = (tip *)malloc(Np * D * sizeof(tip))) == NULL ) fatal ("Nedovoljno memorije_u");

#ifdef ISPIS_U_DAT
	if ((di = fopen (izlazna, "wt")) == NULL) fatal ("Ne mogu otvoriti izlaznu datoteku.");
#endif

	Inicijalizacija(x, dg, gg, D, Np);

#ifndef ISPIS_ZA_GNUPLOT
	ispis_vektora(x, D, Np, G, 'x', di);
#else
	ispis_najbolje_jedinke(x, D, Np, G, ekstrem, di);
#endif

	do {
#ifndef OPTIMIZIRAN_KOD
		Mutacija(x, v, D, Np, F, di);

	#ifndef ISPIS_ZA_GNUPLOT
		ispis_vektora(v, D, Np, G+1, 'v', di);
	#endif

		Rekombinacija(x, v, u, D, Np, CR);

#else
		OPTIMIZIRAN_KOD_Mutacija_i_Rekombinacija(x, u, D, Np, F, CR, di);
#endif

#ifndef ISPIS_ZA_GNUPLOT
		ispis_vektora(u, D, Np, G+1, 'u', di);
#endif

		Selekcija(x, u, D, Np, ekstrem);

		G++;

#ifndef ISPIS_ZA_GNUPLOT
		ispis_vektora(x, D, Np, G, 'x', di);
#else
		ispis_najbolje_jedinke(x, D, Np, G, ekstrem, di);
#endif

	} while (G < stop);

	free(x);
	free(u);
#ifndef OPTIMIZIRAN_KOD
	free(v);
#endif
	free(dg);
	free(gg);

#ifdef ISPIS_U_DAT
	fclose (di);
#endif
	return 0;
}
